from importlib.resources import files

import numpy as np
import pandas as pd
from scipy.constants import R
from scipy.optimize import brentq

from MagmaPandas.EOSs.vinet import Vinet_VdP
from MagmaPandas.parse_io import make_iterable, repeat_vars

_Fe_polymorphs = [
    "Fe_fcc",
    "Fe_bcc-alpha",
    "Fe_HCP",
    "Fe_bcc-delta",
    "Fe_liquid",
]

"""
Polynomial parameters for calculating Gibbs free energy at 1 bar and temperatures above 1000 Kelvin.
Values from Hidayat et al. 2015 for FeO and FeO1.5 and Dinsdale (1991) for O2
"""
# TODO check if data are copied via PyPI
_IW_G0_params_file = (
    files("MagmaPandas.fO2.data").joinpath("IW_G0_params.csv").open("r")
)
_gibbs0_parameters = pd.read_csv(_IW_G0_params_file, index_col=["phase", "temperature"])

# Below 1000 Kelvin for O2:
_O2_low_temperature = pd.Series(
    {
        "a": -6961.7445,  # constant
        "b": -51.0057,  # bT_K
        "c": -22.271,  # cT_Kln(T_K)
        "d": 0,  # d ln(T_K)
        "e": -1.01977e-2,  # dT_K**2
        "f": 1.32369e-8,  # fT_K**3
        "g": -7629.7484,  # g/T_K
        "h": 0.0,  # hT**7
        "i": 0.0,  # iT_K**9
        # "func": Gibbs0_polynomial_1,
    }
)


"""
Values from table Hirschmann et al. (2018) table S2

EOS is from Komabayashi, 2014, with V0 of FeO1.5 taken by extrapolation
of wustite stoichiometry-volume data compiled by Hirschmann et al. (2018, 16.372±0.070)
This is very similar to value from McCammon and Liu 1984 trend (16.425).
Fe parameters from Komabayashi (2014), with bcc the same as fcc, except 
that V0 of bcc is taken from Dorogokupets et al. (2017)

V_0     
    volume at reference conditions
K_0
    bulk modulus at reference conditions
Kprime_0
    first pressure derivative of K_0
alpha0
    thermal expansivity at 1 bar
delta0
    Value of Anderson-Grüneisen parameter (delta_T) at 1 bar
kappa
    dimensionless Anderson-Grüneisen parameter. Set to 1.4 following Komabayashi (2014) and Wood (1993)
"""
# TODO Move these parameters to EOSs.parameters? Make the parameter names uniform, e.g. Kprime_0 vs dKdP
_eos_parameters = pd.DataFrame(
    {
        "V_0": [12.256, 16.372, 6.82, 7.092, 6.753, 7.092, 6.88],  # cm3/mol
        "K_0": [149, 149, 163.4, 163.4, 163.4, 163.4, 148],  # GPa
        "Kprime_0": [3.83, 3.83, 5.38, 5.38, 5.38, 5.38, 5.8],  # GPa-1
        "alpha0": [4.5e-5, 4.5e-5, 7e-05, 7e-05, 5.8e-05, 7e-05, 9e-5],
        "delta0": [4.25, 4.25, 5.5, 5.5, 5.1, 5.5, 5.1],
        "kappa": [1.4, 1.4, 1.4, 1.4, 1.4, 1.4, 1.4],
    },
    index=[
        "FeO",
        "FeO1.5",
        "Fe_fcc",
        "Fe_bcc-alpha",
        "Fe_HCP",
        "Fe_bcc-delta",
        "Fe_liquid",
    ],
)


"""
Mixing parameters in J/mol for FeO-FeO1.5 solid solution from Hidayat et al. (2015)

RT ln gamma(X1) =((q00+2*q10*(1-X2))*X2**2)         eq. S5a
RT ln gamma(X2) =((1-X2)**2*(q00+q10-2*q10*X2))     eq. S5b

XFeO=X1; XFeO1.5=X2
"""
_mixing_parameters = pd.Series({"q00": -5.94e4, "q10": 4.27e4})


def _Gibbs0_polynomial(T_K: float | np.ndarray, a, b, c, d, e, f, g, h, i, **kwargs):

    return (
        a
        + b * T_K
        + c * T_K * np.log(T_K)
        + d * np.log(T_K)
        + e * T_K**2
        + f * T_K**3
        + g / T_K
        + h * T_K**7
        + i * T_K**-9.0
    )


def _Gibbs0(T_K: float | np.ndarray, params: pd.Series | pd.DataFrame):
    """
    calculate Gibbs free energy at 1 bar
    """
    if isinstance(params, pd.DataFrame):
        func = lambda params: _Gibbs0_polynomial(T_K=T_K, **params)
        return params.apply(func, axis=1)

    return _Gibbs0_polynomial(T_K=T_K, **params)


def _Gibbs_Fe_magnetic(T_K):
    """
    Adjust bcc-alpha for magnetic contribution.  Magnetic contribution for fcc
    is << 1 J at temperatures of interest and is ignored

    translated from Hirschmann et al. (2021) matlab script
    """
    Tc = 1043  # curie temperature
    P_factor = 0.4  # structure dependent parameter
    beta = 2.22  # magnetic moment
    A = 1.55828482

    if T_K < Tc:
        gibbs_magnetic = 1 - (1 / A) * (
            (79 * (T_K / Tc) ** (-1) / (140 * P_factor))
            + (474 / 497)
            * ((1 / P_factor) - 1)
            * ((T_K / Tc) ** 3 / 6 + (T_K / Tc) ** 9 / 135 + (T_K / Tc) ** 15 / 600)
        )
    else:
        # why divide T_K by T_K, should this be T_K / Tc?
        gibbs_magnetic = (-1 / A) * (
            (T_K / Tc) ** -5 / 10 + (T_K / Tc) ** -15 / 315 + (T_K / T_K) ** -25 / 1500
        )

    return gibbs_magnetic * (R * T_K * np.log(beta + 1))


def _Gibbs_Fe_polymorphs(P_bar, T_K):
    """
    Calculate Gibbs free energy for Fe polymorphs

    translated from Hirschmann et al. (2021) matlab script
    """
    # TODO there is a tiny difference in pressure adjusted G, check where this comes from.

    P_GPa = P_bar / 1e4
    temperature_range = "high" if T_K > 1811 else "low"
    gibbs0_params = _gibbs0_parameters.loc[(_Fe_polymorphs, temperature_range), :]
    gibbs0 = pd.Series(index=_Fe_polymorphs)
    gibbs_pressure = pd.Series(index=_Fe_polymorphs)

    for ((name, _), G0), (_, eos_params) in zip(
        gibbs0_params.iterrows(), _eos_parameters.loc[_Fe_polymorphs, :].iterrows()
    ):
        gibbs0.loc[name] = _Gibbs0(T_K=T_K, params=G0)
        gibbs_pressure.loc[name] = Vinet_VdP(P_GPa=P_GPa, T_K=T_K, **eos_params)

    gibbs_total = gibbs0.add(gibbs_pressure)
    gibbs_total["Fe_bcc-alpha"] += _Gibbs_Fe_magnetic(T_K=T_K)

    return gibbs_total


def _Gibbs_wustite_O2(P_bar, T_K):
    """
    Calculate gibbs free enery for FeO, FeO1.5 and O2

    translated from Hirschmann et al. (2021) matlab script
    """
    P_GPa = P_bar / 1e4
    temperature_range = "high" if T_K > 1811 else "low"

    gibbs0 = pd.Series(index=["FeO", "FeO1.5", "O2"])
    gibbs_pressure = pd.Series(0.0, index=["FeO", "FeO1.5"])

    gibbs0_params = _gibbs0_parameters.loc[(gibbs0.index, temperature_range), :].copy()
    gibbs0_params.loc["O2"] = (
        _O2_low_temperature.values
        if (T_K < 1000)
        else gibbs0_params.loc[("O2", temperature_range), :].values
    )

    for (name, _), G0 in gibbs0_params.iterrows():
        gibbs0.loc[name] = _Gibbs0(T_K=T_K, params=G0)
    for name, eos_params in _eos_parameters.loc[gibbs_pressure.index, :].iterrows():
        gibbs_pressure.loc[name] = Vinet_VdP(P_GPa=P_GPa, T_K=T_K, **eos_params)

    gibbs_total = gibbs0.add(gibbs_pressure, fill_value=0.0)

    return gibbs_total


def _Gibbs_IW(P_bar, T_K, Fe_phase=False, suppress_Fe_liquid=True):
    """
    Calculate Gibbs free enery for FeO, FeO1.5, O2 and Fe. For Fe, the polymorph with lowest Gibbs free energy is used.

    translated from Hirschmann et al. (2021) matlab script
    """

    gibbs_total = _Gibbs_wustite_O2(P_bar=P_bar, T_K=T_K)

    gibbs_Fe_all = _Gibbs_Fe_polymorphs(P_bar=P_bar, T_K=T_K)
    addendum = ""
    if suppress_Fe_liquid:
        addendum = (
            "*" if gibbs_Fe_all.index[gibbs_Fe_all.argmin()] == "Fe_liquid" else ""
        )
        gibbs_Fe_all = gibbs_Fe_all.drop("Fe_liquid")
    Fe_phase, Fe_gibbs = (
        gibbs_Fe_all.index[gibbs_Fe_all.argmin()],
        gibbs_Fe_all.min(),
    )
    gibbs_total["Fe"] = Fe_gibbs

    output = (gibbs_total, Fe_phase + addendum) if Fe_phase else gibbs_total

    return output


def _deltaGibbs_Fe_wustite(Gibbs: pd.Series):
    """
    Calculate delta Gibbs of the reaction 2 FeO1.5 +  Fe = 3 FeO

    translated from Hirschmann et al. (2021) matlab script

    deltaG : float
        Change of Gibbs free energy of the reaction
    """

    return (3 * Gibbs["FeO"]) - (2 * Gibbs["FeO1.5"]) - Gibbs["Fe"]


def _deltaGibbs_FeOFeO1p5(Gibbs: pd.Series):
    """
    Change in Gibbs free energy of the reaction

    FeO + 1/4 O2 = FeO1.5
    """

    return Gibbs["FeO1.5"] - Gibbs["FeO"] - (Gibbs["O2"] / 4)


def _calculate_XFeO1p5(T_K, deltaGibbs_Fe_wustite, q00, q10, maxiter=300):
    """
    Calculate equilibrium mol fraction FeO1.5 in the reaction

    FeO1.5 + 1/2 Fe = 1.5 FeO

    translated from Hirschmann et al. (2021) matlab script
    """

    func = lambda X_FeO1p5: _Fe_wustite_equilibrium(
        T_K=T_K,
        X_FeO1p5=X_FeO1p5,
        deltaGibbs_Fe_wustite=deltaGibbs_Fe_wustite,
        q00=q00,
        q10=q10,
    )  # Why does Hirschmann square the output? To avoid nagatives?
    try:
        # find solution bracketed by ~0.0 and 1.0
        X_FeO1p5_solution = brentq(
            func, a=1e-6, b=1.0 - 1e-6, xtol=1e-9, maxiter=maxiter
        )
    except RuntimeError as e:
        print(f"{e}\n\nXFeO1.5 did not converge, value set to 1e-6")
        X_FeO1p5_solution = 1e-6
    except ValueError as e:
        print(f"{e}\n\nXFeO1.5 solution outside 0.0-1.0, value set to 1e-6")
        X_FeO1p5_solution = 1e-6

    return X_FeO1p5_solution


def _Fe_wustite_equilibrium(T_K, X_FeO1p5, deltaGibbs_Fe_wustite, q00, q10):
    """
    equilibrium of the reaction

    FeO1.5 + 1/2 Fe = 1.5 FeO (eq. 14)

    as

    deltaG + RT*LN(a_FeO**1.5/aFeO1.5**1) = 0

    including solid solution mixing terms rewritten as

    deltaG + RT*LN(X_FeO**1.5/X_FeO1.5**1) + 1.5 * RT*LN*gammaFeO - RT*LN*gammaFeO1.5

    with:

    RT*LN*gammaFeO          = ((q00+2*q10*(1-X_FeO1.5))*X_FeO1.5**2),
    RT*LN*gammaFeO1.5       = ((1-X_FeO1.5)**2*(q00+q10-2*q10*X_FeO1.5)),
    q00, q10 = solid solution mixing parameters
    a_FeO = gamma * X_FeO = activity of FeO,
    X_FeO1.5 = mol fraction FeO1.5
    X_FeO = 1 - X_FeO1.5

    translated from Hirschmann et al. (2021) matlab script
    """

    RT_LN_gammaFeO = _gamma_FeO(X_FeO1p5=X_FeO1p5, q00=q00, q10=q10)
    RT_LN_gammaFeO1p5 = _gamma_FeO1p5(X_FeO1p5=X_FeO1p5, q00=q00, q10=q10)

    # Half stoichiometry according to Hirschmann et al. (2018) Matlab code
    part_1 = 0.5 * deltaGibbs_Fe_wustite + R * T_K * np.log(
        (1 - X_FeO1p5) ** 1.5 / X_FeO1p5
    )
    part_2 = 1.5 * RT_LN_gammaFeO - RT_LN_gammaFeO1p5

    return part_1 + part_2


def _gamma_FeO(X_FeO1p5, q00, q10):
    """
    equation S5a from Hirschmann et al. (2021)

    RT*LN(a_FeO) =      ((q00+2*q10*(1-X_FeO1.5))*X_FeO1.5**2)

    translated from Hirschmann et al. (2021) matlab script

    Parameters
    ----------
    X_FeO1p5 :  float, array-like
        mol fraction FeO1.5
    q00, q10 : float
        mixing parameters
    """
    return (q00 + 2 * q10 * (1 - X_FeO1p5)) * (X_FeO1p5**2)


def _gamma_FeO1p5(X_FeO1p5, q00, q10):
    """
    equation S5b from Hirschmann et al. (2021)

    RT*LN(a_FeO1.5) =   ((1-X_FeO1.5)**2*(q00+q10-2*q10*X_FeO1.5))

    translated from Hirschmann et al. (2021) matlab script

    Parameters
    ----------
    X_FeO1p5 :  float, array-like
        mol fraction FeO1.5
    q00, q10 : float
        mixing parameters

    """
    return ((1 - X_FeO1p5) ** 2) * (q00 + q10 - 2 * q10 * X_FeO1p5)


@np.vectorize(excluded=["full_output", "suppress_Fe_liquid"])
def _muO2_IW(T_K, P_bar, full_output=False, suppress_Fe_liquid=False):
    """
    calculate chemical potential of oxygen at IW and pressure P with equations of state

    translated from Hirschmann et al. (2021) matlab script
    """

    q00 = _mixing_parameters["q00"]
    q10 = _mixing_parameters["q10"]

    Gibbs, Fe_phase = _Gibbs_IW(
        P_bar=P_bar, T_K=T_K, Fe_phase=True, suppress_Fe_liquid=suppress_Fe_liquid
    )
    deltaG_FeO_FeO1p5 = _deltaGibbs_FeOFeO1p5(Gibbs=Gibbs)
    deltaG_Fe_wustite = _deltaGibbs_Fe_wustite(Gibbs=Gibbs)

    X_FeO1p5 = _calculate_XFeO1p5(
        T_K=T_K,
        deltaGibbs_Fe_wustite=deltaG_Fe_wustite,
        q00=q00,
        q10=q10,
    )

    X_FeO = 1 - X_FeO1p5
    y = 1 - 2 / (X_FeO1p5 + 2)  # Fe(1-y)O

    gamma_1 = _gamma_FeO(X_FeO1p5=X_FeO1p5, q00=q00, q10=q10)
    gamma_2 = _gamma_FeO1p5(X_FeO1p5=X_FeO1p5, q00=q00, q10=q10)

    mu_O2 = 4 * (
        deltaG_FeO_FeO1p5 + R * T_K * np.log(X_FeO1p5 / X_FeO) + gamma_2 - gamma_1
    )  # chemical potential

    output = mu_O2, Fe_phase, y if full_output else mu_O2

    return output


def calculate_fO2(
    logshift: float, T_K, P_bar, full_output=False, suppress_Fe_liquid=False
):
    """
    Calculate oxygen fugacity at the Iron-Wustite buffer according to Hirschmann (2021)\ [18]_

    Parameters
    ----------
    logshift    : int, float
        log units shift of fO2
    T_K         : float, array-like
        temperature in Kelvin
    P_bar       : float, aray-like
        pressure in bar
    full_output : boolean
        ouputs fO2 only if False. fO2, stable Fe phase and Fe(1-y)O if True
    suppress_Fe_liquid  : boolean
        ignore Fe liquid if True


    Returns
    -------
    fO2 : float
        oxygen fugacity
    """
    try:
        logshift = logshift
    except TypeError:
        pass

    offset = 10**logshift

    T_K, P_bar = repeat_vars(T_K, P_bar)

    mu_O2, Fe_phase, y = _muO2_IW(
        T_K=T_K, P_bar=P_bar, full_output=True, suppress_Fe_liquid=suppress_Fe_liquid
    )

    fO2 = np.exp(mu_O2 / (R * T_K)) * offset
    try:
        fO2 = np.float32(fO2.item())
    except ValueError:
        fO2 = fO2.astype(np.float32)

    idx = T_K.index if isinstance(T_K, (pd.Series)) else range(len(mu_O2))
    output = (
        (pd.DataFrame({"fO2": fO2, "Fe_phase": Fe_phase, "Fe(1-y)O": y}, index=idx))
        if full_output
        else fO2
    )

    return output


def _fO2_IW_Campbell(logshift, T_K, P_bar):
    """
    from:
    Campbell et al. (2009) High pressure effects on the iron-iron oxide and nickel-nickel oxide oxygen fugacity buffers. EPSL

    Supplementary table S4

    calibrated between 4-57.6 GPa and 873-2096 K
    """
    P_GPa = P_bar * 1e5 / 1e9
    offset = 10**logshift

    # TODO Armstrong et al. (2018) Supplementary Materials eq. S15 has a positive P**2 coefficient, check which one is correct.
    part_1 = 6.54106 + 1.23e-3 * P_GPa
    part_2 = (-28164 + 546.32 * P_GPa - 1.1341 * P_GPa**2 + 1.93e-3 * P_GPa**3) / T_K

    log10fO2 = part_1 + part_2

    return 10**log10fO2 * offset


def _fO2_FeFeO94_Oneill_Huebner(logshift, T_K, P_bar):
    """
    Fe-FeO(0.94) buffer with the 100 kPa relation from O'neill (1988) and pressure term derived from Huebner (1971).
    Formulation is from Zhang et al. (2017) Supplementary Materials eq. S1
    """

    P_GPa = P_bar * 1e5 / 1e9
    offset = 10**logshift

    part_1 = -28777.89 / T_K + 14.0572
    part_2 = -2.039 * np.log10(T_K) + 550 * (P_GPa - 1e-4) / T_K

    log10fO2 = part_1 + part_2

    return 10 ** (log10fO2) * offset


def _fO2_IW_Zhang(logshift, T_K, P_bar):
    """
    Following Zhang et al. (2017):

    Low pressure part (< 5 GPa) is calculated as an interpolation between Oneill+Huebner and Campbell, while the high pressure part (>5 GPa) is pure Campbell
    """

    P_bar, T_K = make_iterable(P_bar, T_K)

    array_length_one = np.array([len(T_K), len(P_bar)]) == 1
    if array_length_one.any():
        array_length = np.array([len(T_K), len(P_bar)]).max()
        P_bar, T_K = [
            np.ones(array_length) * i[0] if len(i) == 1 else i for i in (P_bar, T_K)
        ]

    low_pressure = P_bar < 5e4
    high_pressure = P_bar > 5e4

    ONeill_Huebner_low_pressure = np.log10(
        _fO2_FeFeO94_Oneill_Huebner(
            logshift=logshift, T_K=T_K[low_pressure], P_bar=P_bar[low_pressure]
        )
    )
    Campbell_low_pressure = np.log10(
        _fO2_IW_Campbell(
            logshift=logshift, T_K=T_K[low_pressure], P_bar=P_bar[low_pressure]
        )
    )

    fO2_low_pressure = 10 ** (
        ONeill_Huebner_low_pressure * (1 - 0.2 * P_bar[low_pressure] / 1e4)
        + (0.2 * P_bar[low_pressure] / 1e4) * Campbell_low_pressure
    )

    fO2_high_pressure = _fO2_IW_Campbell(
        logshift=logshift, T_K=T_K[high_pressure], P_bar=P_bar[high_pressure]
    )

    fO2 = np.concatenate([fO2_low_pressure, fO2_high_pressure])

    return fO2



import inspect
import sys

from MagmaPandas.volatile_solubility.volatile_solubility_models import (
    Allison2022,
    IaconoMarziano,
    Shiskina,
)

_members = inspect.getmembers(sys.modules[__name__], inspect.ismodule)


volatile_solubility_models_dict = {
    cls[0]: cls[1] for cls in _members if cls[1].__package__ == __name__
}



from abc import ABC, abstractmethod


class Solubility_model(ABC):
    @classmethod
    @abstractmethod
    def calculate_saturation(self):
        pass

    @classmethod
    @abstractmethod
    def calculate_solubility(self):
        pass



"""
Module with melt-only and melt-mineral thermometers
"""

import inspect

from MagmaPandas.thermometers import melt, ol_melt

ol_melt_thermometers = {
    func[0]: func[1]
    for func in inspect.getmembers(ol_melt, inspect.isfunction)
    if (func[0][0] != "_") and (func[1].__module__ == ol_melt.__name__)
}



import pandas as pd

from MagmaPandas.Fe_redox.Fe3Fe2_models import Fe3Fe2_models_dict
from MagmaPandas.fO2.fO2_calculate import calculate_fO2
from MagmaPandas.Kd.Ol_melt.FeMg.Kd_models import Kd_olmelt_FeMg_models_dict
from MagmaPandas.magma_protocol import Magma

# configuration needs to be imported last to avoid circular imports
from MagmaPandas.configuration import configuration  # isort: skip


def observed_FeMg_Kd(
    melt: Magma,
    forsterite: pd.Series,
    P_bar,
    T_K: None | float | pd.Series = None,
    **kwargs,
) -> pd.Series:
    """
    Calulate observed Kds based on given melt and olivine compositions.

    Parameters
    ----------
    melt : MagmaFrame
        melt composition in oxide wt. %
    forsterite : float, pandas Series
        olivine forsterite content
    P_bar   : float, pandas Series
        pressures in bar
    T_K : None, float, pandas Series
        temperatures in Kelvin. If set to None, temperatures are calculated according to the melt thermometer set in the global configuration.

    Returns
    -------
    Kds : pandas Series
    """
    if isinstance(P_bar, (int, float)):
        P_bar = pd.Series(P_bar, index=melt.index)
    if isinstance(forsterite, (int, float)):
        forsterite = pd.Series(forsterite, index=melt.index)

    forsterite = forsterite[melt.index]

    melt_x_moles = melt.moles()

    Fe3Fe2_model = Fe3Fe2_models_dict[configuration.Fe3Fe2_model]
    dfO2 = kwargs.get("dfO2", configuration.dfO2)

    if T_K is None:
        T_K = melt_x_moles.temperature(P_bar)
    fO2 = calculate_fO2(T_K=T_K, P_bar=P_bar, dfO2=dfO2)
    Fe3Fe2 = Fe3Fe2_model.calculate_Fe3Fe2(
        melt_mol_fractions=melt_x_moles, T_K=T_K, fO2=fO2, P_bar=P_bar
    )

    melt_x_moles = melt_x_moles.normalise()
    Fe2_FeTotal = 1 / (1 + Fe3Fe2)
    melt_MgFe = melt_x_moles["MgO"] / (melt_x_moles["FeO"] * Fe2_FeTotal)
    olivine_MgFe = forsterite / (1 - forsterite)
    Kd_observed = melt_MgFe / olivine_MgFe

    return Kd_observed


def calculate_FeMg_Kd(
    melt_mol_fractions: pd.Series | pd.DataFrame,
    T_K: float | pd.Series,
    *args,
    **kwargs,
) -> pd.Series:
    """
    Calculate equilibrium Kds for given melt compositions.

    Kd models are set in the global configuration

    Parameters
    ----------
    melt_mol_fractions : pandas Series, pandas Dataframe
        melt composition in oxide mol fractions
    forsterite_initial  : float, pandas Series
        initial forsterite contents. Forsterite values are iteratively adjusted and initial values are not necessarily in Fe-Mg equilibrium with melts.
    T_K : float, pandas Series
        temperatures in Kelvin
    Fe3Fe2  : float, pandas Series
        melt |Fe3Fe2| ratios

    Returns
    -------
    Kds : pandas Series
    """

    Kd_model_name = kwargs.get("Kd_model", configuration.Kd_model)
    Kd_model = Kd_olmelt_FeMg_models_dict[Kd_model_name]

    if Kd_model_name == "blundy":
        dfO2 = kwargs.get("dfO2", configuration.dfO2)
        P_bar = kwargs["P_bar"]
        kwargs["fO2"] = calculate_fO2(T_K=T_K, P_bar=P_bar, dfO2=dfO2)

    return Kd_model.calculate_Kd(
        melt_mol_fractions=melt_mol_fractions, T_K=T_K, *args, **kwargs
    )

    # if isinstance(melt_mol_fractions, pd.Series):
    #     Kd_func = it.calculate_Kd
    # elif isinstance(melt_mol_fractions, pd.DataFrame):
    #     Kd_func = vec.calculate_Kd

    # return Kd_func(
    #     melt_mol_fractions=melt_mol_fractions,
    #     forsterite_initial=forsterite_initial,
    #     T_K=T_K,
    #     Fe3Fe2=Fe3Fe2,
    #     *args,
    #     **kwargs,
    # )



import elementMass as e
import pandas as pd

from MagmaPandas.magma_protocol import Magma
from MagmaPandas.parse_io import check_components

"""
based on Iacovino and Till (2019)\ [23]_
"""

T_reference = pd.Series(
    {
        "SiO2": 1773,
        "TiO2": 1773,
        "Al2O3": 1773,
        "Fe2O3": 1723,
        "FeO": 1723,
        "MgO": 1773,
        "CaO": 1773,
        "Na2O": 1773,
        "K2O": 1773,
        "H2O": 1273,
    }
)

molar_volumes = pd.Series(
    {
        "SiO2": 26.86,
        "TiO2": 28.32,
        "Al2O3": 37.42,
        "Fe2O3": 41.50,
        "FeO": 12.68,
        "MgO": 12.02,
        "CaO": 16.90,
        "Na2O": 29.65,
        "K2O": 47.28,
        "H2O": 22.9,
    }
)

dVdT = pd.Series(
    {
        "SiO2": 0.0,
        "TiO2": 0.00724,
        "Al2O3": 0.00262,
        "Fe2O3": 0.0,
        "FeO": 0.00369,
        "MgO": 0.00327,
        "CaO": 0.00374,
        "Na2O": 0.00768,
        "K2O": 0.01208,
        "H2O": 0.0095,
    }
)

dVdP = pd.Series(
    {
        "SiO2": -1.89e-4,
        "TiO2": -2.31e-4,
        "Al2O3": -2.26e-4,
        "Fe2O3": -2.5e-4,
        "FeO": -4.5e-5,
        "MgO": 2.7e-5,
        "CaO": 3.4e-5,
        "Na2O": -2.4e-4,
        "K2O": -6.75e-4,
        "H2O": -3.2e-4,
    }
)


def calculate_density(
    melt_wt_percent: Magma, T_K: float | pd.Series, P_bar: float | pd.Series
) -> pd.Series:
    """
    Calculate silicate liquid densities according to the model from Iacovino and Till (2019)\ [23]_

    The model uses thermodynamic data from Lange and Carmichael (1987)\ [24]_, Lange (1997)\ [25]_, Kress and Carmichael (1991)\ [2]_ and Ochs III and Lange (1999)\ [26]_

    Parameters
    ----------
    composition : Magma
        melt composition in oxide wt. %
    T_K : float, pandas Series
        temperatures in Kelvin
    P_bar : float, pandas Series
        pressures in bar

    Returns
    -------
    densities : pandas Series
        densities in kg/m\ :sup:`3`
    """
    axis = [0, 1][isinstance(melt_wt_percent, pd.DataFrame)]

    mol_fractions = check_components(melt_wt_percent, molar_volumes.index).moles()[
        molar_volumes.index
    ]

    oxide_masses = e.compound_weights(mol_fractions.columns)

    # equation 1 numerator
    mass_1_mol = mol_fractions.mul(oxide_masses, axis=1).sum(axis=axis)

    T_contribution = _calculate_T_contribution(T_K, index=mol_fractions.index)
    P_contribution = _calculate_P_contribution(P_bar, index=mol_fractions.index)

    V_liquid = (T_contribution + P_contribution).add(
        molar_volumes, axis=1
    ) * mol_fractions

    density = mass_1_mol / V_liquid.sum(axis=axis) * 1e3  # kg.m-3

    return density


def _calculate_temperature(
    melt_wt_percent: Magma, density: pd.Series, P_bar: pd.Series
) -> pd.Series:

    mol_fractions = check_components(
        melt_wt_percent, components=molar_volumes.index
    ).moles()[molar_volumes.index]

    oxide_masses = e.compound_weights(mol_fractions.columns)

    # eauation 1 numerator
    mass_1_mole = mol_fractions.mul(oxide_masses, axis=1).sum(axis=1)

    P_contribution = _calculate_P_contribution(P_bar, index=mol_fractions.index)

    V_liquid = mass_1_mole * 1e3 / density

    V_P_contribution = (P_contribution * mol_fractions).sum(axis=1)
    V_reference = mol_fractions.mul(molar_volumes, axis=1).sum(axis=1)
    temperature_contribution = V_liquid - V_P_contribution - V_reference

    T_contribution_1500K = mol_fractions.mul(
        _calculate_T_contribution(
            pd.Series(1500, index=mol_fractions.index), mol_fractions.index
        ),
        axis=1,
    ).sum(axis=1)
    V_missing = temperature_contribution - T_contribution_1500K
    V_per_degree = mol_fractions.mul(dVdT, axis=1).sum(axis=1)

    return 1500 + (V_missing / V_per_degree)


def _calculate_P_contribution(P_bar: pd.Series, index):
    P_contribution = pd.DataFrame(columns=dVdP.index, index=index)
    P_contribution.loc[:, :] = dVdP.values[None, :]
    P_contribution = P_contribution.mul(P_bar - 1, axis=0)

    return P_contribution


def _calculate_T_contribution(T_K: pd.Series, index):
    if isinstance(T_K, (int, float)):
        T_K = pd.Series(T_K, index=index)

    T_contribution = pd.DataFrame(columns=(T_reference.index), index=index)
    T_contribution.loc[:, :] = T_K.values[:, None]
    T_contribution = T_contribution.sub(T_reference, axis=1)
    T_contribution = T_contribution.mul(dVdT, axis=1)

    return T_contribution



import warnings as w

import numpy as np
from scipy.constants import R

from MagmaPandas.parse_io.validate import _check_value


def fO2_NNO(logshift, T_K, P_bar):
    """
    from:
    Campbell et al. (2009) High pressure effects on the iron-iron oxide and nickel-nickel oxide oxygen fugacity buffers. EPSL

    Supplementary table S5

    calibrated between 2.34-65.9 GPa and 293-2467 K
    """
    P_GPa = P_bar / 1e4
    offset = 10**logshift

    part_1 = (
        8.699 + 1.642e-2 * P_GPa - 3e-4 * P_GPa**2 + 2.7e-6 * P_GPa**3 - 1e-8 * P_GPa**4
    )
    part_2 = (-24205 + 444.73 * P_GPa - 5.929e-1 * P_GPa**2 + 1.53e-3 * P_GPa**3) / T_K

    log10fO2 = part_1 + part_2

    return 10**log10fO2 * offset


@_check_value(var_name="T_K", allowed_range=[700, 1700], error=False)
def fO2_NNO_1bar(
    logshift: float | np.ndarray, T_K: float | np.ndarray
) -> float | np.ndarray:
    """
    Equation 6 from:

    O'Neill and Pownceby (1993) Thermodynamic data from redox reactions at high temperatures. I. An experimental and theoretical assessment of the electrochemical method using stabilized zirconia electrolytes, with revised values for the Fe-“FeO”, Co-CoO, Ni-NiO and Cu-Cu2O oxygen buffers, and new data for the W-WO2 buffer. Contributions to mineralogy and petrology. 114
    """

    try:
        logshift = float(logshift)
    except TypeError:
        pass

    offset = 10**logshift

    muO2 = -478967 + 248.514 * T_K - 9.7961 * np.log(T_K)  # J/mol
    fO2 = np.exp(muO2 / (R * T_K)) * offset

    return fO2


def fO2_NNO_frost(logshift, T_K, P_bar):
    """
    From Table 1 in:

    B. R. Frost (1991) Introduction to oxygen fugacity and its petrological importance. In: Oxide minerals: petrologic and magnetic significance.
    """

    offset = 10**logshift

    log10fO2 = -24930 / T_K + 9.36 + 4.6e-2 * (P_bar - 1) / T_K

    return 10 ** (log10fO2) * offset



import warnings as w
from typing import List, Union

import numpy as np
import pandas as pd


def match_indeces(data: Union[pd.DataFrame, pd.Series], variables: List[pd.Series]):

    for var in variables:

        try:
            if not data.index.equals(var.index):
                raise RuntimeError(f"Data and {var.name} indices don't match")
        except AttributeError:
            pass


def make_iterable(*args):
    return [np.array([a]).flatten() for a in args]


def make_equal_length(*args):
    args_iterable = make_iterable(*args)
    lengths = [len(a) for a in args_iterable]
    max_length = max(lengths)

    if sum([(1 < l < max_length) for l in lengths]) > 0:
        ValueError("Argument has invalid length: 1 < length < max_length")

    return (np.repeat(a, max_length) if len(a) == 1 else a for a in args_iterable)


def convert_to_series(variables: List, index):

    new_vars = []
    for var in variables:
        if isinstance(var, (int, float)):
            new_var = pd.Series(var, index=index)
        else:
            new_var = var.copy()
        new_vars.append(new_var)

    return new_vars


def check_components(composition, components):
    comp = composition.copy()

    if isinstance(comp, pd.DataFrame):
        missing_oxides = set(components).difference(comp.columns)
    elif isinstance(comp, pd.Series):
        missing_oxides = set(components).difference(comp.index)

    if len(missing_oxides) == 0:
        return comp.fillna(0.0)

    for oxide in missing_oxides:
        comp[oxide] = 0.0

    w.warn(
        f"{', '.join(str(i) for i in missing_oxides)} missing in composition and set to 0."
    )

    return comp.fillna(0.0).recalculate()


def repeat_vars(var1, var2):

    var1_is_int, var2_is_int = (
        True if np.issubdtype(type(var), np.number) else False for var in (var1, var2)
    )  # check if var is a float or int (number)
    if bool(var1_is_int) & bool(var2_is_int):
        return make_iterable(var1, var2)
    # If only one variable, P or T, is a single value
    if not (bool(var1_is_int) ^ bool(var2_is_int)):
        return var1, var2

    var1, var2 = (np.array([var]).flatten() for var in (var1, var2))
    # Cycle the short variable
    var1 = np.repeat(var1, len(var2)) if var1_is_int else var1
    var2 = np.repeat(var2, len(var1)) if var2_is_int else var2

    return var1, var2



"""
|CO2|-|H2O| solubility models from Shiskina et al. (2010)\ [34]_
"""

from typing import Tuple

import numpy as np
from scipy.optimize import root, root_scalar

from MagmaPandas.magma_protocol import Magma
from MagmaPandas.parse_io.validate import _check_argument, _check_setter
from MagmaPandas.volatile_solubility.solubility_baseclass import Solubility_model

"""
Equations from:

Shiskina, T. A., Botcharnikov, R. E., Holtz, F., Almeev, R. R., Jazwa, A. M., Jakubiak, A. A. (2014)
Compositional and pressure effects on the solubility of H2Oand CO2 in maficmelt
Chemical geology 388 (112 - 129)

"""
fugacity_options = ["ideal"]


class _meta_shiskina_configuration(type):
    def __init__(cls, *args, **kwargs):
        cls._fugacity = "ideal"

    @property
    def fugacity(cls):
        return cls._fugacity

    @fugacity.setter
    @_check_setter(fugacity_options)
    def fugacity(cls, model: str):
        cls._fugacity = model

    @classmethod
    def reset(cls):
        cls._fugacity = "ideal"
        cls._model = "mixed"


class shiskina_configuration(metaclass=_meta_shiskina_configuration):
    @classmethod
    def reset(cls):
        cls._fugacity = "ideal"

    @classmethod
    def print(cls):
        """ """

        variables = {
            "Fugacity model": "_fugacity",
        }

        pad_left = 20
        pad_right = 20
        pad_total = pad_left + pad_right

        print(" Shiskina volatile solubility ".center(pad_total, "#"))
        print("".ljust(pad_total, "#"))
        print("Settings".ljust(pad_total, "_"))
        for param, model in variables.items():
            print(f"{param:.<{pad_left}}{getattr(cls, model):.>{pad_right}}")
        print("\nCalibration range".ljust(pad_total, "_"))
        T_string = f"1423-1523\N{DEGREE SIGN}K"
        print(f"{'Temperature':.<{pad_left}}{T_string:.>{pad_right}}")
        print(f"{'Pressure':.<{pad_left}}{'0.2-5 kbar':.>{pad_right}}")
        print("\n")


# Model parameters
co2_parameters = {
    "pi": {"A": 1.167, "B": 0.671, "C": 0.65},
    "pi_star": {"A": 1.150, "B": 6.71, "C": -1.345},
}


class h2o(Solubility_model):
    @staticmethod
    def calculate_saturation(oxide_wtPercents: Magma, **kwargs) -> float:
        """
        Calculate melt |H2O| solubility according to equation 9.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %.
        kwargs
            keyword arguments passed to :py:meth:`~MagmaPandas.volatile_solubility.models.Shiskina.h2o.calculate_solubility`

        Returns
        -------
        H2O : float
            melt |H2O| solubility in wt.%.
        """

        if "H2O" not in oxide_wtPercents.index:
            raise ValueError("H2O not found in sample")
        if oxide_wtPercents["H2O"] < 0:
            raise ValueError(f"H2O lower than 0: {oxide_wtPercents['H2O']}")
        if oxide_wtPercents["H2O"] == 0.0:
            return 0.0

        if oxide_wtPercents["H2O"] < h2o.calculate_solubility(
            oxide_wtPercents, P_bar=0
        ):
            return np.nan

        composition = oxide_wtPercents.copy()

        # Upper limit of 15kbar
        upper_limit = 1.5e4
        # try:
        P_saturation = root_scalar(
            h2o._root_saturation,
            args=(composition, kwargs),
            bracket=[1e-15, upper_limit],
        ).root
        # except:
        #     P_saturation = np.nan
        return P_saturation

    @staticmethod
    def calculate_solubility(
        oxide_wtPercents: Magma, P_bar: float, x_fluid: float = 1.0, **kwargs
    ) -> float:
        """
        Calculate melt |H2O| solubility according to equation 9.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %. Needs to have an 'H2O' column.
        P_bar : float
            Pressure in bar.
        x_fluid : float
            fraction of |H2O| in the fluid. Default value is 1.0

        Returns
        -------
        H2O : float
            melt |H2O| solubility in wt.%.
        """
        if not 1 >= x_fluid >= 0:
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")
        if P_bar < 0:
            raise ValueError(f"Pressure is negative: '{P_bar}'")

        oxides = {"K2O", "Na2O"}
        if oxides.difference(oxide_wtPercents.index):
            raise KeyError(
                f"missing oxides: {oxides.difference(oxide_wtPercents.index)}"
            )

        # Calculate cation mol fractions on an anhydrous basis
        volatiles = {"H2O", "CO2"}.intersection(oxide_wtPercents.index)
        if volatiles:
            oxide_wtPercents = oxide_wtPercents.drop(volatiles)
            oxide_wtPercents.recalculate(inplace=True)

        mol_fractions = oxide_wtPercents.cations()

        P_MPa = P_bar / 10

        # Partial pressure (ideal gas fugacity)
        fH2O = x_fluid * P_MPa

        a = 3.36e-7 * fH2O**3 - 2.33e-4 * fH2O**2 + 0.0711 * fH2O - 1.1309
        b = mol_fractions["Na"] + mol_fractions["K"]
        c = -1.2e-5 * fH2O**2 + 0.0196 * fH2O + 1.1297

        return a * b + c

    def _root_saturation(P_bar, oxide_wtPercents, kwargs):

        composition = oxide_wtPercents.copy()

        return composition["H2O"] - h2o.calculate_solubility(
            oxide_wtPercents=composition, P_bar=P_bar, **kwargs
        )


class co2(Solubility_model):
    @staticmethod
    def calculate_saturation(oxide_wtPercents: Magma, **kwargs) -> float:
        """
        Calculate melt |CO2| saturation pressure according to equation 13.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %. Needs to have a 'CO2' column.
        kwargs
            keyword arguments passed to :py:meth:`~MagmaPandas.volatile_solubility.models.Shiskina.co2.calculate_solubility`

        Returns
        -------
        P_saturation : float
            Saturation pressure in bar
        """

        if "CO2" not in oxide_wtPercents.index:
            raise ValueError("CO2 not found in sample")
        if oxide_wtPercents["CO2"] < 0:
            raise ValueError(f"CO2 lower than 0: {oxide_wtPercents['H2O']}")
        if oxide_wtPercents["CO2"] == 0.0:
            return 0.0

        composition = oxide_wtPercents.copy()

        # Upper limit of 15kbar
        upper_limit = 1.5e4
        # try:
        P_saturation = root_scalar(
            co2._root_saturation,
            args=(composition, kwargs),
            bracket=[1e-15, upper_limit],
        ).root
        # except:
        #     P_saturation = np.nan
        return P_saturation

    @staticmethod
    def calculate_solubility(
        oxide_wtPercents: Magma, P_bar: float, x_fluid: float = 0.0, **kwargs
    ) -> float:
        """
        Calculate melt |CO2| solubility according to equation 13.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %.
        P_bar : float, array-like
            Pressure in bar
        x_fluid : float
            fraction of |H2O| in the fluid. Default value is 0.0

        Returns
        -------
        solublity : float
            melt |CO2| solubility in wt. %.
        """
        if not 1 >= x_fluid >= 0:
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")
        if P_bar < 0:
            raise ValueError(f"Pressure is negative: '{P_bar}'")
        if x_fluid == 1.0:
            return 0.0

        P_MPa = P_bar / 10
        # Partial pressure (ideal gas fugacity)
        fCO2 = (1 - x_fluid) * P_MPa

        model = kwargs.get("model", "pi_star")
        A = co2_parameters[model]["A"]
        B = co2_parameters[model]["B"]
        C = co2_parameters[model]["C"]

        mol_fractions = oxide_wtPercents.cations()
        pi_star = co2._pi_star(mol_fractions)

        return np.exp(A * np.log(fCO2) + B * pi_star + C) / 1e4

    @staticmethod
    def _root_saturation(P_bar, oxide_wtPercents, kwargs):
        """ """
        composition = oxide_wtPercents.copy()

        return composition["CO2"] - co2.calculate_solubility(
            oxide_wtPercents=composition, P_bar=P_bar, **kwargs
        )

    @staticmethod
    def _pi_star(mol_fractions):
        """
        equation 11
        """
        cations = {"Ca", "K", "Na", "Mg", "Fe", "Si", "Al"}
        if cations.difference(mol_fractions.index):
            raise KeyError(
                f"missing cations: {cations.difference(mol_fractions.index)}"
            )

        a = (
            mol_fractions["Ca"]
            + 0.8 * mol_fractions["K"]
            + 0.7 * mol_fractions["Na"]
            + 0.4 * mol_fractions["Mg"]
            + 0.4 * mol_fractions["Fe"]
        )
        b = mol_fractions["Si"] + mol_fractions["Al"]

        return a / b


class mixed(Solubility_model):
    @staticmethod
    @_check_argument("output", [None, "PXfl", "P", "Xfl"])
    def calculate_saturation(
        oxide_wtPercents: Magma, output: str = "P", **kwargs
    ) -> float | Tuple[float, float]:
        """
        Calculate volatile saturation pressure for systems with mixed |CO2|-|H2O| fluids.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composision in oxide wt. %. Needs to have 'H2O' and 'CO2' columns.
        output : str
            Output format. 'P' for pressure only, 'Xfl' for |H2O| fluid fraction only and 'PXfl' for both.

        Returns
        -------
        saturation : float, (float, float)
            Depending on the value of ``output``: saturation pressure in bar, |H2O| fluid fraction or (saturation pressure, fluid fraction)
        """

        composition = oxide_wtPercents.copy()

        P_H2O_saturation = h2o.calculate_saturation(composition, x_fluid=1.0)
        P_CO2_saturation = co2.calculate_saturation(composition, x_fluid=0.0)

        if oxide_wtPercents["H2O"] <= 0:
            return P_CO2_saturation
        if oxide_wtPercents["CO2"] <= 0:
            return P_H2O_saturation

        P_guess = 0

        for species in (P_H2O_saturation, P_CO2_saturation):
            if np.isfinite(species):
                P_guess += species

        saturation = root(
            mixed._saturation_rootFunction,
            x0=[P_guess, 0.01],
            args=(composition),
        ).x

        if saturation[1] <= 0.0:
            saturation[0] = P_CO2_saturation
        elif saturation[1] >= 1.0:
            saturation[0] = P_H2O_saturation
        saturation[1] = np.clip(saturation[1], 0.0, 1.0)

        return_dict = {"P": saturation[0], "Xfl": saturation[1], "PXfl": saturation}
        return return_dict[output]

    @staticmethod
    @_check_argument("output", [None, "both", "CO2", "H2O"])
    def calculate_solubility(
        oxide_wtPercents: Magma,
        P_bar: float,
        x_fluid: float,
        output: str = "both",
        **kwargs,
    ) -> float | Tuple[float, float]:
        """
        Calculate volatile solubilities for systems with mixed |CO2|-|H2O| fluids.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composision in oxide wt. %.
        P_bar : float
            pressure in bar
        x_fluid: float
            fraction of |H2O| in the fluid.
        output : str
            Output format. 'CO2' for |CO2| only, 'H2O' for |H2O| only and 'both' for both.

        Returns
        -------
        saturation : float, (float, float)
            Solubility in wt. %. Depending on the value of ``output``: |CO2|, |H2O| or (|CO2|, |H2O|).
        """

        if not 1 >= x_fluid >= 0:
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")
        if P_bar < 0:
            raise ValueError(f"Pressure is negative: '{P_bar}'")

        composition = oxide_wtPercents.copy()
        H2O = h2o.calculate_solubility(composition, P_bar, x_fluid)
        composition["H2O"] = np.float32(H2O)
        CO2 = co2.calculate_solubility(composition, P_bar, x_fluid)

        return_dict = {"both": (H2O, CO2), "H2O": H2O, "CO2": CO2}
        return return_dict[output]

    @staticmethod
    def _saturation_rootFunction(P_x_fluid, oxide_wtPercents):
        """
        compare calculated with sample concentrations
        """

        P_bar, x_fluid = P_x_fluid
        # Keep x_fluid and P_bar within bounds
        x_fluid = np.clip(x_fluid, 0.0, 1.0)
        P_bar = np.clip(P_bar, a_min=1e-15, a_max=None)

        composition = oxide_wtPercents.copy()

        H2O = composition["H2O"]
        CO2 = composition["CO2"]

        sample_concentrations = np.array([H2O, CO2])
        calculated_concentrations = np.array(
            mixed.calculate_solubility(
                oxide_wtPercents=composition,
                P_bar=P_bar,
                x_fluid=x_fluid,
                output="both",
            )
        )

        return abs(calculated_concentrations - sample_concentrations)



import re
from typing import Dict, List

import elementMass as e
import numpy as np
import pandas as pd
from typing_extensions import Self

from MagmaPandas.configuration import configuration
from MagmaPandas.Elements import element_weights, oxide_compositions
from MagmaPandas.enums import Datatype, Unit
from MagmaPandas.parse_io.validate import _check_argument, _check_attribute
from MagmaPandas.thermometers.melt import melt_thermometers_dict


def _MagmaSeries_expanddim(data=None, *args, **kwargs):
    from MagmaPandas.MagmaFrames import MagmaFrame

    if isinstance(data, MagmaSeries):
        kwargs["units"] = data._units
        kwargs["datatype"] = data._datatype
        kwargs["weights"] = data._weights.copy(deep=True)

    df = MagmaFrame(data, *args, **kwargs)

    # df._units = data._units.copy(deep=True)
    # df._datatype = data._datatype.copy(deep=True)

    return df


# # pd.concat (pandas/core/reshape/concat.py) requires this for the
# # concatenation of series since pandas 1.1
# # (https://github.com/pandas-dev/pandas/commit/f9e4c8c84bcef987973f2624cc2932394c171c8c)
# _MagmaSeries_expanddim._get_axis_number = pd.DataFrame._get_axis_number


class MagmaSeries(pd.Series):
    """
    Generic MagmaSeries class for chemical data.

    Parameters
    ----------
    data : array-like, Iterable, dict, or scalar value
        geochemical data with elements or oxides as index
    units : None, str
        data units, either "mol fraction", "wt. %" or "ppm"
    datatype : None, str
        datatype either "cation" or "oxide"
    weights : None, pandas Series
        atomic weights of elements or oxides in the MagmaSeries
    """

    # New attributes
    _metadata = ["_weights", "_units", "_datatype"]

    @_check_argument("units", [None, "mol fraction", "wt. %", "ppm"])
    @_check_argument("datatype", [None, "cation", "oxide"])
    def __init__(
        self,
        data=None,
        *args,
        units: str = None,
        datatype: str = None,
        weights: pd.Series = None,
        **kwargs,
    ) -> None:
        self._units: Unit = Unit(units)
        self._datatype: Datatype = Datatype(datatype)

        super().__init__(data, *args, **kwargs)

        if weights is not None:
            self._weights = weights.copy(deep=True)
        elif not hasattr(self, "_weights"):
            self._weights = element_weights.weights_as_series(self.index)

    @property
    def _constructor(self):
        """
        This is the key to letting Pandas know how to keep
        derivatives of `MagmaBase` the same type as yours.  It should
        be enough to return the name of the Class.  However, in
        some cases, `__finalize__` is not called and `new attributes` are
        not carried over (see: https://github.com/pandas-dev/pandas/issues/13208).
        We can fix that by constructing a callable
        that makes sure to call `__finalize__` every time.
        """

        def _c(*args, **kwargs):
            if (weights := getattr(self, "_weights", None)) is not None:
                weights = weights.copy(deep=True)

            # The finalize call crashes the Python kernel when using a dictionary as data:
            # 'Cannot execute code, session has been disposed. Please try restarting the Kernel.'
            # WHY?
            return MagmaSeries(*args, weights=weights, **kwargs)  # .__finalize__(self)

        return _c

    @property
    def _constructor_expanddim(self):
        def _c(*args, **kwargs):
            if (weights := getattr(self, "_weights", None)) is not None:
                weights = weights.copy(deep=True)

            return _MagmaSeries_expanddim(
                *args, weights=weights, **kwargs
            ).__finalize__(self)

        # pd.concat (pandas/core/reshape/concat.py) requires this for the
        # concatenation of series since pandas 1.1
        # (https://github.com/pandas-dev/pandas/commit/f9e4c8c84bcef987973f2624cc2932394c171c8c)
        _c._get_axis_number = pd.DataFrame._get_axis_number

        return _c

    @property
    def _no_data(self) -> List:
        """
        Names of all index without chemical data
        """
        no_data = list(self.index.difference(self.elements))
        if "total" in no_data:
            no_data.remove("total")
        return no_data

    @property
    def _total(self) -> bool:
        """
        Dataframe contains totals in index
        """
        if "total" in self.index:
            return True
        else:
            return False

    @property
    def units(self) -> str:
        """
        Datatype and units.
        """
        return f"{self._datatype.value} {self._units.value}"

    @units.setter
    def units(self, value):
        print("units are read only")

    @property
    def weights(self) -> pd.Series:
        """
        Atomic weights of all elements in the MagmaSeries.
        """
        return self._weights

    @property
    def elements(self) -> List[str]:
        """
        Names of all elements in the MagmaSeries
        """
        return list(self._weights.index)

    def moles(self, normalise=True) -> Self:
        """
        Data converted to mol fractions.
        """
        if self._units == Unit.MOL_FRACTIONS:
            return self.copy()

        if self._units == Unit.WT_PERCENT:
            return self._convert_moles_wtPercent(normalise=normalise)
        elif self._units == Unit.PPM:
            return self.convert_ppm_wtPercent()._convert_moles_wtPercent(
                normalise=normalise
            )

        return self.copy()

    def wt_pc(self, normalise=True) -> Self:
        """
        Data converted to wt. %.
        """
        if self._units == Unit.WT_PERCENT:
            return self.copy()

        if self._units == Unit.MOL_FRACTIONS:
            return self._convert_moles_wtPercent(normalise=normalise)
        elif self._units == Unit.PPM:
            return self.convert_ppm_wtPercent()

        return self.copy()

    @property
    def ppm(self) -> Self:

        if self._units == Unit.PPM:
            return self.copy()

        if self._units == Unit.WT_PERCENT:
            return self.convert_ppm_wtPercent()
        elif self._units == Unit.MOL_FRACTIONS:
            return self._convert_moles_wtPercent().convert_ppm_wtPercent()

        return self.copy()

    def cations(self, normalise=True, norm_to=1, mol_fractions=True) -> Self:
        """
        Data converted to cation mol fractions
        """
        if (self._datatype == Datatype.CATION) & (
            mol_fractions & (self._units == Unit.MOL_FRACTIONS)
        ):
            return self.copy()

        moles = self.moles(normalise=False)[self.elements]

        cations_per_oxide = oxide_compositions.cation_amount(moles.elements)
        cations = moles.mul(pd.Series(cations_per_oxide, index=moles.elements))
        # Rename index to cations
        cations.index = oxide_compositions.cation_names(cations.elements)

        cations._datatype = Datatype.CATION
        cations = cations.recalculate()

        if not mol_fractions:
            cations = cations[cations.elements].mul(cations.weights)
            cations._units = Unit.WT_PERCENT
            norm_to = 100

        if not normalise:
            cations["total"] = cations.sum(axis=1)
            return cations.recalculate()

        # Normalise to 1
        total = cations.sum()
        cations = cations.div(total) * norm_to
        cations["total"] = norm_to
        # Set the right datatype and elements

        cations.recalculate(inplace=True)

        return cations

    def oxides(self, normalise=True, oxidation_state: Dict[str, int] = {}) -> Self:
        """
        Data converted to oxides
        """

        if (self._datatype == Datatype.OXIDE) & (not bool(oxidation_state)):
            return self.copy()

        units = self._units

        cations = self[self.elements].cations(
            normalise=False
        )  # convert to cation mol fractions
        cation_names = cations.elements
        cation_element_names = [
            re.sub(r"\d+", "", e) for e in cation_names
        ]  # strip numbers/charges from the names

        cation_names_new = [
            (
                i
                if oxidation_state.get(j, None) is None
                else f"{j}{int(oxidation_state[j])}"
            )
            for i, j in zip(cation_names, cation_element_names)
        ]  # new names include oxidation state for elements with non-default values

        oxide_names = e.get_oxide_names(cation_names_new)
        cations_per_oxide = e.cation_numbers(oxide_names)

        oxides = cations.rename(
            index={cation: oxide for cation, oxide in zip(cation_names, oxide_names)}
        ).recalculate()  # rename to oxides

        oxides = oxides.div(cations_per_oxide)  # recalculate to oxides
        oxides["total"] = oxides[oxides.elements].sum()
        oxides._datatype = Datatype.OXIDE

        if units == Unit.MOL_FRACTIONS:
            if not normalise:
                return oxides
            return oxides.normalise()

        oxides_wt_pc = oxides.wt_pc(normalise=False)  # recalculate to wt. %
        if not normalise:
            return oxides_wt_pc
        return oxides_wt_pc.normalise()

    def _convert_moles_wtPercent(self, normalise=True) -> Self:
        """
        moles converted to wt. % and vice versa
        """

        converted = self.copy()[self.elements]
        if self._units == Unit.WT_PERCENT:
            converted = converted.div(converted.weights)
            units = Unit.MOL_FRACTIONS
        elif self._units == Unit.MOL_FRACTIONS:
            converted = converted.mul(converted.weights)
            units = Unit.WT_PERCENT

        if not normalise:
            converted["total"] = converted.sum()
            converted._units = units
            return converted

        # Normalise
        total = converted.sum()
        converted = converted.div(total)
        converted["total"] = converted.sum()
        # Set the right units

        if self._units == Unit.MOL_FRACTIONS:
            converted = converted.mul(100)

        converted._units = units

        return converted

    def convert_ppm_wtPercent(self) -> Self:
        """
        ppm converted to wt. % and vice versa
        """
        convert_dict = {
            Unit.WT_PERCENT: [1e4, Unit.PPM],
            Unit.PPM: [1e-4, Unit.WT_PERCENT],
        }

        converted = self.mul(convert_dict[self._units][0])
        converted._units = convert_dict[self._units][1]

        return converted

    def mineral_formula(self, O: int = None) -> Self:
        """
        Calculate mineral formulas by normalising to oxygen per formula unit

        Parameters
        ----------
        O : int
            Amount of oxygen to normalise to.

        Returns
        -------
        mineral formulas : MagmaSeries
        """
        # Calculate cation fractions
        cations = self.cations()
        cations = cations[cations.elements]
        # Calculate oxygens per cation
        oxygen_numbers = e.oxygen_numbers(self.elements) / e.cation_numbers(
            self.elements
        )
        oxygen_numbers.index = cations.elements
        # Normalise to oxygen
        oxygen_total = cations.mul(oxygen_numbers).sum()
        oxygen_factor = O / oxygen_total
        cations = cations.mul(oxygen_factor)
        cations["O"] = O

        return cations

    def recalculate(self, inplace=False) -> Self:
        """
        Recalculate element masses and total weight.
        """

        series = self if inplace else self.copy()

        series._weights = element_weights.weights_as_series(series.index)

        if series._total:
            series["total"] = series[series.elements].sum()

        if not inplace:
            return series

    def normalise(self, to=None) -> Self:
        """
        Normalise compositions.

        Parameters
        ----------
        to :    float, int
            normalisation value

        Returns
        -------
        normalised data : MagmaSeries
        """
        if to is not None:
            norm = to
        elif self._units == Unit.WT_PERCENT:
            norm = 100
        else:
            norm = 1

        normalised = self.recalculate()[self.elements]

        total = normalised.sum()
        normalised = normalised.div(total)
        normalised = normalised.mul(norm)
        normalised["total"] = normalised.sum()

        return normalised

    def temperature(self, *args, **kwargs) -> float:
        """
        Calculate melt liquidus temperature.
        Thermometer models are selected in the global configuration.

        Parameters
        ----------
        P_bar   : float, pandas Series
            pressure in bar

        Returns
        -------
        temperatures : float
            Liquidus temperature in Kelvin
        """

        thermometer = melt_thermometers_dict[configuration.melt_thermometer]

        return thermometer(melt=self.wt_pc(), *args, **kwargs)

    @_check_argument("total_Fe", ["FeO", "Fe2O3"])
    def FeO_Fe2O3_calc(
        self,
        Fe3Fe2: float,
        total_Fe: str = "FeO",
        inplace: bool = False,
        wtpc=True,
    ) -> Self:
        """
        Calculate melt FeO and |Fe2O3| based on total Fe.

        Parameters
        ----------
        Fe3Fe2 : pandas Series
            melt |Fe3Fe2| ratios
        total_Fe    : str
            columname in Melt frame with total Fe
        inplace : bool

        Returns
        -------
        Melt    : Self
            melt compositions inclusding FeO and |Fe2O3|
        """

        Fe2Fe_total = 1 / (1 + Fe3Fe2)
        melt_mol_fractions = self.moles()

        if total_Fe == "FeO":
            Fe2 = melt_mol_fractions["FeO"] * Fe2Fe_total
            Fe3 = melt_mol_fractions["FeO"] * (1 - Fe2Fe_total) / 2
        if total_Fe == "Fe2O3":
            Fe2 = melt_mol_fractions["Fe2O3"] * Fe2Fe_total * 2
            Fe3 = melt_mol_fractions["Fe2O3"] * (1 - Fe2Fe_total)

        melt_mol_fractions["FeO"] = Fe2
        melt_mol_fractions["Fe2O3"] = Fe3
        melt_mol_fractions.recalculate(inplace=True)

        # Recalculate to wt. % (normalised)
        melt = melt_mol_fractions.wt_pc() if wtpc else melt_mol_fractions

        if inplace:
            self["FeO"] = melt["FeO"]
            self["Fe2O3"] = melt["Fe2O3"]
            self.recalculate(inplace=True)

        else:
            return melt

    def random_sample(self, errors) -> Self:
        """
        Randomly resample compositions within errors.

        Sampling distribution is assumed normal with measured values as means and errors as standard deviations.

        Parameters
        ----------
        errors  : float, array-like
            standard deviation of the normal distributions. Use int for a fixed value for all elements or an array for specific values for all elements in :py:attr:`~MagmaPandas.MagmaFrames.magmaFrame.MagmaFrame.elements`

        Returns
        -------
        resampled data : MagmaSeries
            Randomly resampled compositions
        """

        random_sample = np.random.normal(self[self.elements], errors)
        random_sample[random_sample < 0] = 0

        sr = self.copy()
        sr[sr.elements] = random_sample

        return sr



from typing import Union

from MagmaPandas.configuration import configuration
from MagmaPandas.volatile_solubility import volatile_solubility_models as vsm
from MagmaPandas.volatile_solubility.solubility_baseclass import Solubility_model


def get_solubility_model(
    model_name: Union[str, None] = None, species: Union[str, None] = None
) -> Solubility_model:

    if model_name is None:
        model_name = configuration.volatile_solubility
    model = getattr(vsm, model_name)
    if species is None:
        species = configuration.volatile_species

    return getattr(model, species)


def calculate_saturation(
    oxide_wtPercents,
    *args,
    solubility_model: Union[Solubility_model, None] = None,
    **kwargs,
):
    """
    Docstring
    """
    if solubility_model is None:
        solubility_model = get_solubility_model()

    equation = solubility_model.calculate_saturation

    return equation(oxide_wtPercents, *args, **kwargs)


def calculate_solubility(
    oxide_wtPercents,
    P_bar,
    *args,
    solubility_model: Union[Solubility_model, None] = None,
    **kwargs,
):
    """
    Docstring
    """
    if solubility_model is None:
        solubility_model = get_solubility_model()

    equation = solubility_model.calculate_solubility

    return equation(oxide_wtPercents, P_bar, *args, **kwargs)



from typing import Callable

import pandas as pd

from MagmaPandas.parse_io import convert_to_series, match_indeces


def equilibrium_forsterite(melt_mol_fractions, Kd, Fe3Fe2):
    """
    Parameters
    ----------
    melt_mol_fractions  :
        melt composition on mol fractions.
    Kd  :
        (Fe2+/Mg)olivine / (Fe2+/Mg)liquid partitioning coefficient
    Fe3Fe2  :
        melt Fe3+/Fe2+ ratio

    Returns
    -------
    Equilibrium forsterite fraction as Mg/(Mg + Fe)
    """

    # Mg_no = 1 + (1 + Fe2Mg_liquid)
    Fe2FeTotal = 1 / (1 + Fe3Fe2)
    Fe2Mg_liquid = melt_mol_fractions["FeO"] * Fe2FeTotal / melt_mol_fractions["MgO"]

    return 1 / (1 + Kd * Fe2Mg_liquid)


def iterate_Kd_scalar(
    melt_mol_fractions: pd.Series,
    forsterite_initial,
    T_K,
    Fe3Fe2,
    Kd_model: Callable,
    *args,
    **kwargs,
):
    """

    Parameters
    ----------
    melt_mol_fractions : pd.DataFrame
        melt composition in oxide mol fraction.
    olivine_forsterite
        forsterite fraction in olivine as Mg * 100 / (Mg + Fe)
    T_K :
        Temperature in Kelvin
    P_bar :
        Pressure in bar
    Fe3Fe2 :
        melt Fe3+/Fe2+ ratio
    """
    # model_name = kwargs.pop("model", configuration.Kd_model)
    # Kd_model = Kd_models[model_name]

    fo_converge_default = 0.001
    fo_converge = kwargs.pop("fo_converge", fo_converge_default)

    for var, name in zip(
        (forsterite_initial, T_K, Fe3Fe2), ("forsterite_initial", "T_K", "Fe3Fe2")
    ):
        if not isinstance(var, (float, int)):
            raise ValueError(f"{name} should be float or int")

    # initialise Kd
    Kd = Kd_model(
        melt_mol_fractions=melt_mol_fractions,
        forsterite=forsterite_initial,
        T_K=T_K,
        **kwargs,
    )

    forsterite_EQ = equilibrium_forsterite(
        melt_mol_fractions=melt_mol_fractions, Kd=Kd, Fe3Fe2=Fe3Fe2
    )

    # Difference between observed Fo and equilibrium Fo
    forsterite_delta = abs(forsterite_initial - forsterite_EQ) / forsterite_initial

    iterate = forsterite_delta > fo_converge
    # iterate until equilibrium forsterite content doesn't change any more
    while iterate:

        Kd = Kd_model(
            melt_mol_fractions=melt_mol_fractions,
            forsterite=forsterite_initial,
            T_K=T_K,
            **kwargs,
        )

        forsterite_initial = forsterite_EQ.copy()
        forsterite_EQ = equilibrium_forsterite(
            melt_mol_fractions=melt_mol_fractions, Kd=Kd, Fe3Fe2=Fe3Fe2
        )
        forsterite_delta = abs(forsterite_initial - forsterite_EQ) / forsterite_initial
        iterate = forsterite_delta > fo_converge

    return Kd


def iterate_Kd_vectorized(
    melt_mol_fractions: pd.DataFrame,
    forsterite_initial,
    T_K,
    Fe3Fe2,
    Kd_model: Callable,
    **kwargs,
):
    """

    Parameters
    ----------
    melt_mol_fractions : pd.DataFrame
        melt composition in oxide mol fraction.
    olivine_forsterite
        forsterite fraction in olivine as Mg * 100 / (Mg + Fe)
    T_K :
        Temperature in Kelvin
    P_bar :
        Pressure in bar
    Fe3Fe2 :
        melt Fe3+/Fe2+ ratio
    """

    # model_name = kwargs.pop("model", configuration.Kd_model)
    # Kd_model = Kd_models[model_name]

    match_indeces(melt_mol_fractions, [T_K, Fe3Fe2, *kwargs.values()])

    melt_mol_fractions = melt_mol_fractions.copy()

    forsterite, T_K, Fe3Fe2, *args = convert_to_series(
        [forsterite_initial, T_K, Fe3Fe2, *kwargs.values()], melt_mol_fractions.index
    )
    # if len(kwargs.keys()) < 2:
    #     args = [args]

    kwargs = {key: value for key, value in zip(kwargs.keys(), args)}

    fo_converge_default = 0.001
    fo_converge = kwargs.pop("fo_converge", fo_converge_default)

    # initialise Kds
    Kd = Kd_model(
        melt_mol_fractions=melt_mol_fractions, forsterite=forsterite, T_K=T_K, **kwargs
    )

    forsterite_EQ = equilibrium_forsterite(
        melt_mol_fractions=melt_mol_fractions, Kd=Kd, Fe3Fe2=Fe3Fe2
    )
    # Difference between observed Fo and equilibrium Fo
    forsterite_delta = abs(forsterite - forsterite_EQ) / forsterite

    iterate = forsterite_delta > fo_converge
    # iterate until equilibrium forsterite content doesn't change any more
    while sum(iterate) > 0:

        Kd = Kd_model(
            melt_mol_fractions=melt_mol_fractions,
            forsterite=forsterite_EQ,
            T_K=T_K,
            **kwargs,
        )
        forsterite[iterate] = forsterite_EQ[iterate].copy()
        forsterite_EQ = equilibrium_forsterite(
            melt_mol_fractions=melt_mol_fractions, Kd=Kd, Fe3Fe2=Fe3Fe2
        )

        forsterite_delta = abs(forsterite - forsterite_EQ) / forsterite
        iterate = forsterite_delta > fo_converge

    return Kd



import pandas as pd
from typing_extensions import Self

from MagmaPandas.MagmaFrames.magmaFrame import MagmaFrame


class Plagioclase(MagmaFrame):
    """
    Subclass of :py:class:`~MagmaPandas.MagmaFrames.magmaFrame.MagmaFrame` extended with plagioclase specific methods.
    """

    # @property
    # def _constructor(self):
    #     """This is the key to letting Pandas know how to keep
    #     derivatives of `MagmaBase` the same type as yours.  It should
    #     be enough to return the name of the Class.  However, in
    #     some cases, `__finalize__` is not called and `new attributes` are
    #     not carried over.  We can fix that by constructing a callable
    #     that makes sure to call `__finalize__` every time."""

    #     def _c(*args, weights=None, **kwargs):
    #         if weights is None:
    #             weights = self._weights.copy(deep=True)
    #         return Plagioclase(*args, weights=weights, **kwargs).__finalize__(self)

    #     return _c

    @property
    def anorthite(self) -> pd.Series:
        """
        Anorthite contents.
        """
        cations = self.cations()
        return pd.Series(
            cations["Ca"] * 100 / (cations["Ca"] + cations["Na"]), name="An"
        )

    @property
    def endmembers(self) -> pd.DataFrame:
        """
        endmember componenents
        """

        cations = self.cations()

        anorthite = cations["Ca"] * 100 / (cations["Ca"] + cations["Na"] + cations["K"])
        albite = cations["Na"] * 100 / (cations["Ca"] + cations["Na"] + cations["K"])
        orthoclase = cations["K"] * 100 / (cations["Ca"] + cations["Na"] + cations["K"])

        return pd.DataFrame(
            {"anorthite": anorthite, "albite": albite, "orthoclase": orthoclase}
        )

    @property
    def formula(self) -> Self:
        """
        Mineral formulas normalised to 8 O p.f.u.
        """

        return self.mineral_formula(O=8)
